#ifndef EXPERIMENTS_TEST_HARNESS_HPP
#define EXPERIMENTS_TEST_HARNESS_HPP

#include <cmath>
#include <cstdlib>
#include <ctime>
#include <fstream>
#include <iostream>
#include <vector>
#include <adept.h>
#include <adolc/adouble.h>
#include <adolc/drivers/drivers.h>
#include <adolc/drivers/taylor.h>
#include <adolc/taping.h>
#include <adolc/interfaces.h>
#include <cppad/cppad.hpp>
#include <cppad/example/cppad_eigen.hpp>
#include <Eigen/Dense>  // has to come after CppAD's Eigen
#include <stan/model/model_header.hpp>
// #include <stan/math/rev.hpp>
#include <Sacado.hpp>

template <typename F>
void adept_gradient(const F& f, const Eigen::VectorXd& x, double& fx,
                    Eigen::VectorXd& grad_fx) {
  Eigen::Matrix<adept::adouble, Eigen::Dynamic, 1> x_ad(x.size());
  for (int i = 0; i < x.size(); ++i)
    x_ad(i) = x(i);
  adept::active_stack()->new_recording();
  adept::adouble fx_ad = f(x_ad);
  fx = fx_ad.value();
  fx_ad.set_gradient(1.0);
  adept::active_stack()->compute_adjoint();
  grad_fx.resize(x.size());
  for (int i = 0; i < x.size(); ++i)
    grad_fx(i) = x_ad[i].get_gradient();
}

template <typename F>
void adolc_gradient(const F& f, const Eigen::VectorXd& x, double& fx,
                    Eigen::VectorXd& grad_fx) {
  grad_fx.resize(x.size());
  trace_on(1);
  Eigen::Matrix<adouble, Eigen::Dynamic, 1> x_ad(x.size());
  for (int n = 0; n < x.size(); ++n)
    x_ad(n) <<= x(n);
  adouble y = f(x_ad);
  y >>= fx;
  trace_off();
  gradient(1, x.size(), &x(0), &grad_fx(0));
}

template <typename F>
void cppad_gradient(const F& f, const Eigen::VectorXd& x, double& fx,
                    Eigen::VectorXd& grad_fx) {
  Eigen::Matrix<CppAD::AD<double>, Eigen::Dynamic, 1> x_ad(x.size());
  for (int n = 0; n < x.size(); ++n)
    x_ad(n) = x[n];
  Eigen::Matrix<CppAD::AD<double>, Eigen::Dynamic, 1> y(1);
  CppAD::Independent(x_ad);
  y[0] = f(x_ad);
  CppAD::ADFun<double> g = CppAD::ADFun<double>(x_ad, y);
  fx = Value(y[0]);
  Eigen::VectorXd w(1);
  w(0) = 1.0;
  grad_fx =  g.Reverse(1, w);
}

template <typename F>
void sacado_gradient(const F& f,
                     const Eigen::VectorXd& x,
                     double& fx,
                     Eigen::VectorXd& grad_fx) {
  Eigen::Matrix<Sacado::Rad::ADvar<double>, Eigen::Dynamic, 1> x_ad(x.size());
  for (int n = 0; n < x.size(); ++n)
    x_ad(n) = x[n];
  fx = f(x_ad).val();
  Sacado::Rad::ADvar<double>::Gradcomp();

  grad_fx.resize(x.size());
  for (int n = 0; n < x.size(); ++n)
    grad_fx(n) = x_ad(n).adj();
}


void print_results_header(std::ostream& os) {
  os << "model, system, calls, seconds, size\n";
}

inline double elapsed_secs(const clock_t& start) {
  return static_cast<double>(clock() - start) / CLOCKS_PER_SEC;
}

const int NUM_CALLS = 1000;

inline void print_result(const clock_t& start, const std::string& model,
                         const char* system, int size,
                         std::ostream& os) {
  clock_t end = clock();
  double seconds = elapsed_secs(start);
  os << model << "," 
     << system << "," 
     << NUM_CALLS << "," 
     << seconds << "," 
     << size << "\n";
}

template <typename F>
inline void time_gradients(const F& f, const Eigen::VectorXd& x, 
                           std::ostream& os) {
  int N = x.size();
  Eigen::VectorXd grad_fx(N);
  double fx = 0;
  clock_t start;
  std::string f_name = F::name();
  double z = 0;

  z = 0;
  start = clock();
  for (int m = 0; m < NUM_CALLS; ++m) {
    adept_gradient(f, x, fx, grad_fx);
    z += fx;
  }
  print_result(start, F::name(), "adept", N, os);

  z = 0;
  start = clock();
  for (int m = 0; m < NUM_CALLS; ++m) {
    adolc_gradient(f, x, fx, grad_fx);
    z += fx;
  }
  print_result(start, F::name(), "adolc", N, os);

  z = 0;
  start = clock();
  for (int m = 0; m < NUM_CALLS; ++m) {
    cppad_gradient(f, x, fx, grad_fx);
    z += fx;
  }
  print_result(start, F::name(), "cppad", N, os);

  z = 0;
  start = clock();
  for (int m = 0; m < NUM_CALLS; ++m) {
    sacado_gradient(f, x, fx, grad_fx);
    z += fx;
  }
  print_result(start, F::name(), "sacado", N, os);

  z = 0;
  start = clock();
  for (int m = 0; m < NUM_CALLS; ++m) {
    stan::agrad::gradient(f, x, fx, grad_fx);
    z += fx;
  }
  print_result(start, F::name(), "stan", N, os);

  z = 0;
  start = clock();
  for (int m = 0; m < NUM_CALLS; ++m)
    z += f(x);
  print_result(start, F::name(), "double", N, os);
  std::cout << "z=" << z << std::endl;
}


template <typename F>
inline void run_test() {
  adept::Stack adept_global_stack_;
  F f;
  std::string file_name = F::name() + "_eval.csv";
  std::fstream fs(file_name, std::fstream::out);
  print_results_header(fs);
  int max = 4 * 1024; 
  for (int N = 1; N <= max; N *= 2) {
    std::cout << "N = " << N << std::endl;
    Eigen::VectorXd x(N);
    F::fill(x);
    time_gradients(f, x, fs);
  }
  fs.close();
}

#endif
